"""Kernels for SiFT filtering."""

# Note: this module is strongly inspired by the kernel module of sklearn
# <https://github.com/scikit-learn/scikit-learn/blob/main/sklearn/gaussian_process/kernels.py>
#

import numbers
from abc import ABC, abstractmethod
from typing import Any, Optional, Union

import numpy as np
import pykeops
from pykeops.common.lazy_tensor import GenericLazyTensor
from scipy.sparse import issparse

import sift._backend_utils as bu
import sift._types as st
from sift._constants import Backend, KernelType

__all__ = [
    "MappingKernel",
    "PrecomputedKernel",
    "KnnKernel",
    "RBFKernel",
    "LaplacianKernel",
]
torch = bu._safe_import("torch", raise_exc=False)


def _num_samples(x):
    """Return number of samples in array-like x."""
    message = "Expected sequence or array-like, got %s" % type(x)

    if not hasattr(x, "__len__") and not hasattr(x, "shape"):
        if hasattr(x, "__array__"):
            x = np.asarray(x)
        else:
            raise TypeError(message)

    if hasattr(x, "shape") and x.shape is not None:
        if len(x.shape) == 0:
            raise TypeError(
                "Singleton array %r cannot be considered a valid collection." % x
            )
        # Check that shape is returning an integer or default to len
        # Dask dataframes may not return numeric shape[0] value
        if isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]

    try:
        return len(x)
    except TypeError as type_error:
        raise TypeError(message) from type_error


class Kernel(ABC):
    """Base kernel class."""

    def __init__(
        self,
        kernel: Optional[st.ArrayLike] = None,
        dtype: Optional[st.DtypeLike] = None,
        device: Optional[st.Device_t] = None,
        backend: Optional[Backend] = None,
    ):
        self._k = kernel
        self._dtype = dtype
        self._backend = bu.get_backend(backend)
        self._device = bu.get_device(self._backend) if device is None else device

    def __repr__(self):
        return f"{self.__class__.__name__}"

    @abstractmethod
    def __call__(
        self,
        x: st.ArrayLike,
        y: Optional[st.ArrayLike] = None,
        materialize: bool = False,
    ) -> st.KernelLike:
        """Evaluate the kernel on a pair of inputs.

        Parameters
        ----------
        x
            Left argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
        y
            Right argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
            If `None`, :math:`k(X, X)` is evaluated instead.
        materialize
            Whether to materialize the kernel.

        Returns
        -------
        LazyTensor resembling the kernel.
        """

    @classmethod
    def _create(cls, kernel_type: st.Kernel_t, **kwargs: Any) -> "Kernel":
        kernel_type = KernelType(kernel_type)
        if kernel_type == KernelType.PRECOMPUTED:
            return PrecomputedKernel(**kwargs)
        if kernel_type == KernelType.KNN:
            return KnnKernel(**kwargs)
        if kernel_type == KernelType.MAPPING:
            return MappingKernel(**kwargs)
        if kernel_type == KernelType.RBF:
            return RBFKernel(**kwargs)
        if kernel_type == KernelType.LAPLACIAN:
            return LaplacianKernel(**kwargs)
        raise NotImplementedError(kernel_type)

    @property
    def dtype(self) -> st.DtypeLike:
        """The kernel's data type."""
        return self._dtype

    @dtype.setter
    def dtype(self, dtype: st.DtypeLike) -> None:
        self._dtype = dtype

    @property
    def device(self) -> Optional[st.Device_t]:
        """The kernel's device."""
        return self._device

    @device.setter
    def device(self, device: Optional[st.Device_t]) -> None:
        self._device = device

    @property
    def backend(self) -> Backend:
        """The kernel's :mod:`pykeops` backend.

        Can be either :mod:`torch` or :mod:`numpy`."""
        return self._backend

    @backend.setter
    def backend(self, backend: Union[str, Backend]) -> None:
        self._backend = Backend(backend)

    @property
    def k(self) -> st.KernelLike:
        """The instantiated kernel object."""
        if self._k is None:
            raise RuntimeError(
                "Must instantiate a kernel object or supply a valid basis"
            )
        return self._k

    @k.setter
    def k(self, kernel: st.KernelLike) -> None:
        self._k = kernel

    def _materialize_kernel(self):
        if isinstance(self.k, GenericLazyTensor):
            if self.backend == Backend.TORCH:
                return self.k @ torch.diag(
                    torch.ones(self.k.shape[1], dtype=self.dtype, device=self.device)
                )
            elif self.backend == Backend.NUMPY:
                return self.k @ np.diag(np.ones(self.k.shape[1]))
        else:
            return self.k

    def _LazyTensor(self, x, axis=0):
        if self._backend == Backend.TORCH:
            if not torch.is_tensor(x):
                x = torch.tensor(x, dtype=self.dtype, device=self.device).contiguous()
            if axis == 0:
                return pykeops.torch.LazyTensor(x[:, None, None])
            return pykeops.torch.LazyTensor(x[None, :, None])
        elif self._backend == Backend.NUMPY:
            return pykeops.numpy.LazyTensor(x)

    def _Vi(self, x):
        if self._backend == Backend.TORCH:
            if not torch.is_tensor(x):
                x = torch.tensor(x, dtype=self.dtype, device=self.device).contiguous()
            return pykeops.torch.Vi(x)
        elif self._backend == Backend.NUMPY:
            return pykeops.numpy.Vi(x)

    def _Vj(self, x):
        if self._backend == Backend.TORCH:
            if not torch.is_tensor(x):
                x = torch.tensor(x, dtype=self.dtype, device=self.device).contiguous()
            return pykeops.torch.Vj(x)
        elif self._backend == Backend.NUMPY:
            return pykeops.numpy.Vj(x)

    def _Pm(self, x):
        if self._backend == Backend.TORCH:
            if not torch.is_tensor(x):
                x = torch.tensor(x, dtype=self.dtype, device=self.device).contiguous()
            return pykeops.torch.Pm(x)
        elif self._backend == Backend.NUMPY:
            return pykeops.numpy.Pm(x)

    def _normalize(self, K) -> GenericLazyTensor:
        """Row normalize an array.

        Parameters
        ----------
        K
            The array to normalize.

        Returns
        -------
        The row-normalized kernel.
        """
        norm = K.sum(1)
        # norm = np.where(norm == 0)
        return self._Vi(1 / norm) * K


class ConstantKernel(Kernel):
    r"""Constant kernel.

    Apply row-normalization to the provided constant.

    .. math::
        k(x_1, x_2) = constant\_value \;\forall\; x_1, x_2

    Parameters
    ----------
    constant_value
        The constant value which defines the covariance k(x_1, x_2) = constant_value
    kwargs
        Keyword arguments for the parent class.
    """

    def __init__(self, constant_value: float = 1.0, **kwargs: Any):
        super().__init__(**kwargs)
        self.constant_value = constant_value

    def __call__(
        self,
        x: st.ArrayLike,
        y: Optional[st.ArrayLike] = None,
        materialize: bool = False,
    ) -> st.KernelLike:
        """Compute the constant kernel.

        Parameters
        ----------
        x
            Left argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
        y
            Right argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
            If `None`, :math:`k(X, X)` is evaluated instead.
        materialize
            Whether to materialize the kernel.

        Returns
        -------
        The kernel.
        """

        if y is None:
            y = x
        if self.backend == Backend.TORCH:
            self.k = torch.full(
                (_num_samples(x), _num_samples(y)),
                self.constant_value,
                dtype=self.dtype,
                device=self.device,
            )
        elif self.backend == Backend.NUMPY:
            self.k = np.full(
                (_num_samples(x), _num_samples(y)),
                self.constant_value,
            )
        if materialize:
            return self._materialize_kernel()
        return self.k

    def __repr__(self) -> str:
        return f"{self.constant_value:.3g}"


class PrecomputedKernel(Kernel):
    """Precomputed kernel.

    Defines a kernel based on a given pre-computed similarity and
    applies row-normalization to the provided kernel.
    """

    def __call__(
        self,
        x: st.ArrayLike,
        y: Optional[st.ArrayLike] = None,
        materialize: bool = False,
    ) -> st.KernelLike:
        """Return the row-normalized kernel.

        Parameters
        ----------
        x
            Array of shape `[n_samples, n_samples]` to normalize.
        y
            Ignored, used for API compatibility.
        materialize
            Ignored, the kernel is already pre-computed.

        Returns
        -------
        The kernel.
        """

        if issparse(x):
            x = x.copy()
            row_sums = np.array(x.sum(axis=1))[:, 0]
            row_indices, col_indices = x.nonzero()
            x.data[:] = x.data / row_sums[row_indices]
            k = x
        elif self.backend == Backend.TORCH:
            if not torch.is_tensor(x):
                x = torch.tensor(x, dtype=self.dtype, device=self.device).contiguous()
            k = x / torch.sum(x, 1).unsqueeze(-1)
        elif self.backend == Backend.NUMPY:
            k = x / np.sum(x, 1)[:, np.newaxis]
        else:
            raise ValueError(
                f"received {self.backend} which is not a valid SiFT backend. See sift.Backends."
            )
        self.k = k
        return k

    def __repr__(self) -> str:
        return "K"


class KnnKernel(PrecomputedKernel):
    """k-NN kernel.

    Defines a kernel based on a given pre-computed k-nn similarity and
    applies row-normalization to the provided kernel.
    """

    def __repr__(self) -> str:
        return "knn"


class MappingKernel(Kernel):
    r"""Mapping kernel.

    Define a kernel based on a given mapping :math:`T`.

    .. math::
        k(x_i, x_j) =  \sum_k p_{l}(k) p_{c}(k)

    where :math:`p_{l} = T / T.sum(0), p_{c} = T / T.sum(1)`

    Parameters
    ----------
    ignore_self
        Whether to ignore self transitions.
    kwargs
        Keyword arguments for the parent class.
    """

    def __init__(self, ignore_self: bool = False, **kwargs: Any):
        super().__init__(**kwargs)
        self._ignore_self = ignore_self

    def __call__(
        self,
        x: st.ArrayLike,
        y: Optional[st.ArrayLike] = None,
        materialize: bool = False,
    ) -> st.KernelLike:
        """Compute the mapping kernel.

        Parameters
        ----------
        x
            Left argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
        y
            Right argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
            If `None`, :math:`k(X, X)` is evaluated instead.
        materialize
            Whether to materialize the kernel.

        Returns
        -------
        The kernel.
        """
        if y is None:
            y = x
        if x is y:
            y = y.copy()

        if issparse(x) and issparse(y):
            row_sums = np.array(x.sum(axis=1))[:, 0]
            row_indices, col_indices = x.nonzero()
            x.data[:] = x.data / row_sums[row_indices]
            pl = x

            y = y.T
            col_sums = np.array(y.sum(axis=1))[:, 0]
            row_indices, col_indices = y.nonzero()
            y.data[:] = y.data / col_sums[row_indices]
            pc = y.T

            K = pl @ pc.T
            if self._ignore_self:
                K.setdiag(0)
                K = self._normalize(K)
            self.k = K

        elif self.backend == Backend.TORCH:
            if not torch.is_tensor(x):
                x = torch.tensor(x, dtype=self.dtype, device=self.device).contiguous()
            pl = x / torch.sum(x, 1).unsqueeze(-1)

            if not torch.is_tensor(y):
                y = torch.tensor(y, dtype=self.dtype, device=self.device).contiguous()
            pc = y / torch.sum(y, 0)
            K = self._Vi(pl) | self._Vj(pc)
            if self._ignore_self:
                i = self._LazyTensor(np.arange(x.shape[0]), axis=0)
                j = self._LazyTensor(np.arange(y.shape[0]), axis=1)
                K = K - K * (0.5 - (i - j) ** 2).step()
                K = self._normalize(K)
            self.k = K
        elif self.backend == Backend.NUMPY:
            pl = x / np.sum(x, 1)[:, np.newaxis]
            pc = y / np.sum(y, 1)
            K = pl @ pc.T

            if self._ignore_self:
                np.fill_diagonal(K, 0)
                K = self._normalize(K)
            self.k = K
        else:
            raise ValueError(
                f"received {self.backend} which is not a valid SiFT backend."
            )

        if materialize:
            return self._materialize_kernel()

        return self.k

    def __repr__(self) -> str:
        return "Mapping kernel, T"


class RBFKernel(Kernel):
    """Radial basis function kernel (aka squared-exponential kernel).

    The RBFKernel kernel is a stationary kernel. It is also known as the
    "squared exponential" kernel. It is parameterized by a length scale
    parameter :math:`l>0`, which can either be a scalar (isotropic variant
    of the kernel) or a vector with the same number of dimensions as the inputs
    X (anisotropic variant of the kernel). The kernel is given by:

    .. math::
        k(x_i, x_j) = \\exp\\left(- \\frac{d(x_i, x_j)^2}{2l^2} \\right)

    where :math:`l` is the length scale of the kernel and :math:`d(x_i,x_j)`
    is the Euclidean distance.

    For advice on how to set the length scale parameter, see e.g. :cite:`duvenaud:14`, Chapter 2.
    This kernel is infinitely differentiable, which implies that GPs with this
    kernel as covariance function have mean square derivatives of all orders,
    and are thus very smooth. See :cite:`rasmussen:05`, Chapter 4, Section 4.2, for further details.

    Parameters
    ----------
    length_scale
        The length scale.
    ignore_self
        Whether to ignore self transitions.
    kwargs
        Keyword arguments for the parent class.
    """

    def __init__(
        self, length_scale: float = 1.0, ignore_self: bool = False, **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.length_scale = length_scale
        self._ignore_self = ignore_self

    def __call__(
        self,
        x: st.ArrayLike,
        y: Optional[st.ArrayLike] = None,
        materialize: bool = False,
    ) -> st.KernelLike:
        """Compute the RBF kernel.

        Parameters
        ----------
        x
            Left argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
        y
            Right argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
            If `None`, :math:`k(X, X)` is evaluated instead.
        materialize
            Whether to materialize the kernel.

        Returns
        -------
        The kernel.
        """

        if y is None:
            y = x

        K = (
            -self._Pm(1 / (2 * self.length_scale**2))
            * self._Vi(x).sqdist(self._Vj(y))
        ).exp()
        if self._ignore_self:
            i = self._LazyTensor(np.arange(x.shape[0]), axis=0)
            j = self._LazyTensor(np.arange(y.shape[0]), axis=1)
            K = K - K * (0.5 - (i - j) ** 2).step()
        self.k = self._normalize(K)

        if materialize:
            return self._materialize_kernel()

        return self.k

    def __repr__(self) -> str:
        ls = np.ravel(self.length_scale)[0]
        return f"{self.__class__.__name__}(length_scale={ls:.3g})"


class LaplacianKernel(Kernel):
    r"""L1-exponential kernel.

    It is parameterized by a length scale parameter :math:`l>0`,
    The kernel is given by:

    .. math::
        k(x_i, x_j) = \exp\left(- \frac{| x_i, x_j|_{1}}{2l^2} \right)

    where :math:`l` is the length scale of the kernel.

    Parameters
    ----------
    length_scale
        The length scale.
    ignore_self
        Whether to ignore self transitions.
    kwargs
        Keyword arguments for the parent class.
    """

    def __init__(
        self, length_scale: float = 1.0, ignore_self: bool = False, **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.length_scale = length_scale
        self._ignore_self = ignore_self

    def __call__(
        self,
        x: st.ArrayLike,
        y: Optional[st.ArrayLike] = None,
        materialize: bool = False,
    ) -> st.KernelLike:
        """Compute the Laplacian kernel.

        Note that this compound kernel returns the results of all simple kernel
        stacked along an additional axis.

        Parameters
        ----------
        x
            Left argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
        y
            Right argument of the returned kernel :math:`k(X, Y)`, array of shape ``[n_samples, n_features]``.
            If `None`, :math:`k(X, X)` is evaluated instead.
        materialize
            Whether to materialize the kernel.

        Returns
        -------
        The kernel.
        """
        if y is None:
            y = x

        K = (
            -(
                self._Pm(1 / (2 * self.length_scale**2))
                * self._Vi(x).sqdist(self._Vj(y))
            ).sqrt()
        ).exp()

        if self._ignore_self:
            i = self._LazyTensor(np.arange(x.shape[0]), axis=0)
            j = self._LazyTensor(np.arange(y.shape[0]), axis=1)
            K = K - K * (0.5 - (i - j) ** 2).step()
        self.k = self._normalize(K)

        if materialize:
            return self._materialize_kernel()
        return self.k

    def __repr__(self) -> str:
        ls = np.ravel(self.length_scale)[0]
        return f"{self.__class__.__name__}(length_scale={ls:.3g})"



from typing import Literal, Union

import numpy as np
import scipy.sparse as sp

__all__ = ["ArrayLike", "DtypeLike", "Device_t", "Kernel_t"]


ArrayLike = Union[np.ndarray, sp.spmatrix]
KernelLike = np.ndarray
DtypeLike = np.dtype

Device_t = Literal["cpu", "cuda"]
try:
    import torch

    Device_t = Union[Device_t, torch.device]
    ArrayLike = Union[ArrayLike, torch.Tensor]
    DtypeLike = Union[DtypeLike, torch.dtype]
    KernelLike = Union[KernelLike, torch.Tensor]
except ImportError:
    pass

Kernel_t = Literal["precomputed", "knn", "mapping", "rbf", "laplacian"]



import logging
import os
from types import MappingProxyType
from typing import Any, Dict, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from anndata import AnnData
from scipy.sparse import issparse

import sift._backend_utils as bu
import sift._types as st
from sift._constants import BATCH_SIZE, MAX_SRC_SIZE, Backend, KernelType, Stage
from sift._utils import (
    _get_embedding,
    _get_mask_idx,
    _prepare_kernel_container,
    _save_res,
)
from sift._utils import plot_kernel as _plot_kernel
from sift.kernels import Kernel

logger = logging.getLogger(__name__)


__all__ = ["SiFT", "sifter"]


class SiFT:
    """SiFT object.

    Parameters
    ----------
    adata
        Annotated data object.
    kernel_key
        Key in :class:`anndata.AnnData` where information for distance is stored.
    metric
        The kernel type.
    src_key
        Mask for the source space of the kernel.
    tgt_key
        Mask for the target space of the kernel.
    knn_key
        wey in :attr:`anndata.AnnData.obs` for k-NN masking.
    save: bool
        Whether to save the kernel.
    kernel_params
        Kernel object parameters [length_scale].
    kwargs
        Keyword arguments for kernel computations  [precomputed_kernel, n_neighbors, knn_batch_key].

    Examples
    --------
    >>> import scanpy as sc
    >>> import sift
    >>> adata = sc.read(...)
    >>> sft = sift.SiFT(
            adata=adata,
            kernel_key="cell_cycle_genes",
            metric="RBF",
        )
    >>> sft.filter()
    """

    def __init__(
        self,
        adata: AnnData,
        kernel_key: Optional[str] = None,
        metric: st.Kernel_t = KernelType.RBF,
        mask_key: Optional[str] = None,
        src_key: Optional[str] = None,
        tgt_key: Optional[str] = None,
        copy: bool = False,
        kernel_params: Optional[Dict[str, Any]] = MappingProxyType({}),
        **kwargs: Any,
    ):
        metric = KernelType(metric)
        self._adata = adata.copy() if copy else adata
        self._kernel_key = kernel_key
        self._kernel_metric = metric
        self._src_key = src_key
        self._tgt_key = tgt_key
        self._stage = Stage.INITIALIZED
        self._kernel = None
        self._kernel_mask = None
        self._mask_key = None
        self._kernel_containers = {}
        self._copy = copy

        backend = bu.get_backend(None)
        bu.assert_backend_available(backend)
        device = bu.get_device(backend)
        dtype = bu.get_dtype(backend)

        self._device = kwargs.pop("device", device)
        self._dtype = kwargs.pop("dtype", dtype)
        self._backend = kwargs.pop("backend", backend)
        self._batch_size = kwargs.pop("batch_size", BATCH_SIZE)
        self._max_src_size = kwargs.pop("max_src_size", MAX_SRC_SIZE)

        self._init_kernels(
            kernel_key=kernel_key,
            mask_key=mask_key,
            metric=metric,
            kernel_params=kernel_params,
            **kwargs,
        )

        self._model_summary_string = f"SiFT object with {metric} kernel"
        logger.info(f"initialized a SiFTer with {metric} kernel.")
        if self._mask_key is not None:
            logger.info(f"SiFTer has a mask_kernel of type {self._mask_key}")

    @property
    def adata(self) -> AnnData:
        """The SiFT annotated data object."""
        return self._adata

    @adata.setter
    def adata(self, adata: AnnData) -> None:
        self._adata = adata

    @property
    def kernel(self) -> st.KernelLike:
        """The :mod:`pykeops` kernel object."""
        return self._kernel(**self._kernel_containers)

    @kernel.setter
    def kernel(self, k: st.KernelLike) -> None:
        self._kernel = k

    def _init_kernels(
        self,
        kernel_key: Optional[str] = None,
        mask_key: Optional[str] = None,
        metric: st.Kernel_t = KernelType.RBF,
        precomputed_kernel: Optional[st.ArrayLike] = None,
        kernel_params: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Sets the cell-cell similarity kernel.

        Parameters
        ----------
        kernel_key
            Key in :class:`anndata.AnnData` where information for distance is stored.
        mask_key
            Key in :attr:`anndata.AnnData.obs` for k-NN masking.
        metric
          The kernel type.
        precomputed_kernel
            The user can provide a pre-computed kernel, it is assumed cells are ordered as in :attr:`adata`.
        kernel_params
            Kernel object parameters [length_scale].
        kwargs
            Keyword arguments for kernel computations  [precomputed_kernel, n_neighbors, knn_batch_key].

        Returns
        -------
        Nothing, just initializes the kernels.
        """
        (
            self._kernel_containers["x"],
            self._kernel_containers["y"],
        ) = _prepare_kernel_container(
            adata=self.adata,
            kernel_key=kernel_key,
            metric=metric,
            precomputed_kernel=precomputed_kernel,
            src_key=self._src_key,
            tgt_key=self._tgt_key,
            **kwargs,
        )

        self._kernel = Kernel._create(
            metric,
            device=self._device,
            dtype=self._dtype,
            backend=self._backend,
            **kernel_params,
        )

        if mask_key is not None:
            x, y = _prepare_kernel_container(
                adata=self.adata,
                kernel_key=mask_key,
                metric=KernelType.KNN,
                src_key=self._src_key,
                tgt_key=self._tgt_key,
                uns_key="mask_kernel",
                **kwargs,
            )

            _kernel_mask = Kernel._create(
                KernelType.MAPPING,
                device=self._device,
                dtype=self._dtype,
                backend=self._backend,
            )
            _kernel_mask(x=x, y=y)

            self._kernel_mask = (
                _kernel_mask.k.A if issparse(_kernel_mask.k) else _kernel_mask.k
            )
            self._mask_key = mask_key

    def _materialize_kernel(
        self,
        k: Optional[st.ArrayLike] = None,
        to_cpu: Optional[bool] = True,
    ) -> st.KernelLike:
        """Materialize the initialized kernel.

        Parameters
        ----------
        k
            Kernel to materialize, if `None` call `self._kernel.kernel()`.
        to_cpu: bool
            Whether to move kernel to CPU.

        Returns
        -------
        The materialized kernel object.

        Notes
        -----
        This operation can result in out-of-memory (OOM) error.
        """
        k = (
            k
            if k is not None
            else self._kernel(materialize=True, **self._kernel_containers)
        )
        k = k.cpu().numpy() if to_cpu and bu.is_torch_tensor(k) else k
        k = k.A if issparse(k) else k
        return k

    def materialize_kernel(
        self, kernel_containers: Optional[Dict] = None, to_cpu: bool = True
    ) -> st.KernelLike:
        """Materialize the initialized kernel.

        Parameters
        ----------
        to_cpu
            Whether to move kernel to CPU.
        kernel_containers
            Keyword arguments to the kernel call.

        Returns
        -------
        The materialized kernel after masking, if :attr:`._kernel_mask` is present.

        Notes
        -----
        This operation can result in out-of-memory (OOM) error.
        """
        kernel_containers = (
            kernel_containers
            if kernel_containers is not None
            else self._kernel_containers
        )
        if self._mask_key is not None:
            return self._materialize_kernel(self._mask_kernel(), to_cpu=to_cpu)
        else:
            return self._materialize_kernel(
                self._kernel(materialize=True, **kernel_containers), to_cpu=to_cpu
            )

    def _mask_kernel(
        self, kernel_input_idx: Optional[Dict[str, Any]] = None
    ) -> st.ArrayLike:
        """Mask the kernel.

        Parameters
        ----------
        kernel_input_idx
            Dictionary with index for kernel input.

        Returns
        -------
        The materialized kernel after masking, if :attr:`_kernel_mask` is present.

        Notes
        -----
        This operation can result in out-of-memory (OOM) error.
        """
        if kernel_input_idx is not None:
            x = self._kernel_containers["x"][kernel_input_idx["x"], :]
            y = self._kernel_containers["y"][kernel_input_idx["y"], :]
            kernel_mask = (
                self._kernel_mask[kernel_input_idx["x"], kernel_input_idx["y"]]
                if self._mask_key
                else None
            )
        else:
            x = self._kernel_containers["x"]
            y = self._kernel_containers["y"]
            kernel_mask = self._kernel_mask if self._mask_key else None

        if self._mask_key is not None:
            if self._backend == Backend.TORCH and not issparse(self.kernel):
                k = (
                    self._materialize_kernel(
                        k=self._kernel(x=x, y=y, materialize=True), to_cpu=False
                    )
                    * kernel_mask
                )
            elif issparse(self.kernel) and self._mask_key is not None:
                k = self.kernel.multiply(kernel_mask)
            else:
                k = (
                    self._materialize_kernel(
                        k=self._kernel(materialize=True, **self._kernel_containers)
                    )
                    * kernel_mask
                )
        else:
            k = self._kernel(x=x, y=y)
        return k

    def filter(
        self,
        embedding_key: Optional[str] = None,
        use_raw: Optional[bool] = False,
        pseudocount: bool = False,
        key_added: Optional[str] = None,
    ) -> Optional[AnnData]:
        """Perform filtering, subtract the projection of the kernel from the expression.

        Parameters
        ----------
        embedding_key
            Key in :attr:`anndata.AnnData.obsm` for which to compute the projection on.
            If `None`, use :attr:`anndata.AnnData.X`.
        use_raw
            Whether to use :attr:`anndata.AnnData.raw`, if present.
        pseudocount
            If `True`, add a pseudocount to filtered expression to avoid negative values
        key_added
            Key used when saving the result. If `None`, use ``'{embedding_key}_sift'``.
            The result is saved to :attr:`anndata.AnnData.X`, :attr:`anndata.AnnData.layers` or
            :attr:`anndata.AnnData.obsm`, depending on the ``embedding_key``.

        Returns
        -------
        The filtered data matrix.
        """
        if use_raw is True and self.adata.raw is None:
            raise ValueError("Received `use_raw=True`, but `adata.raw` is empty.")
        if use_raw and self.adata.raw is not None:
            logger.info("Using `adata.raw`")
            self.adata = self.adata.raw.to_adata()

        emb, type_emb, embedding_key = _get_embedding(
            adata=self.adata, embedding_key=embedding_key
        )
        tgt_idx = (
            _get_mask_idx(adata=self.adata, key=self._tgt_key)
            if self._tgt_key is not None
            else np.arange(self.adata.n_obs)
        )

        src_idx = (
            _get_mask_idx(adata=self.adata, key=self._src_key)
            if self._src_key is not None
            else np.arange(self.adata.n_obs)
        )

        logger.info(
            f"Filtering cell-cell similarity kernel using projection on `{embedding_key}`."
        )
        if self._backend == Backend.TORCH:
            import torch

            if issparse(self.kernel):
                k = self._mask_kernel()
                emb_sifted = emb[src_idx, :] - k @ emb[tgt_idx, :]
            else:
                emb_src = emb[src_idx, :]
                if src_idx.shape[0] > self._max_src_size:
                    batches = np.array_split(
                        np.arange(src_idx.shape[0]),
                        np.ceil(src_idx.shape[0] / self._batch_size),
                    )
                    logger.info(f"Evaluating with {len(batches)} batches.")
                    y_idx = np.arange(tgt_idx.shape[0])
                    emb_sifted_batches = []
                    emb_tgt = torch.tensor(
                        emb[tgt_idx, :].A if issparse(emb) else emb[tgt_idx, :],
                        dtype=self._dtype,
                        device=self._device,
                    )
                    for batch in batches:
                        k = self._mask_kernel(kernel_input_idx={"x": batch, "y": y_idx})
                        emb_src_batch = torch.tensor(
                            emb_src[batch, :].A if issparse(emb) else emb_src[batch, :],
                            dtype=self._dtype,
                            device=self._device,
                        )
                        emb_sifted_batch = emb_src_batch - k @ emb_tgt
                        emb_sifted_batches.append(emb_sifted_batch.cpu().numpy())
                    emb_sifted = np.concatenate(emb_sifted_batches)
                else:
                    k = self._mask_kernel()
                    emb = torch.tensor(
                        emb.A if issparse(emb) else emb,
                        dtype=self._dtype,
                        device=self._device,
                    )
                    emb_sifted = emb[src_idx, :] - k @ emb[tgt_idx, :]

        else:  # numpy backend
            k = self._mask_kernel()
            emb = (
                emb.astype(dtype=self.kernel.dtype).A
                if issparse(emb)
                else emb.astype(dtype=self.kernel.dtype)
            )
            if issparse(self.kernel):
                emb_sifted = emb[src_idx, :] - k @ emb[tgt_idx, :]
            else:
                emb_sifted = emb[src_idx, :] - k.__matmul__(
                    emb[tgt_idx, :], backend="CPU"
                )

        if pseudocount:
            emb_sifted -= emb_sifted.min()

        if bu.is_torch_tensor(emb_sifted):
            emb_sifted = emb_sifted.cpu().numpy()

        emb_sifted_full = np.zeros((self.adata.n_obs, emb_sifted.shape[-1]))
        emb_sifted_full[src_idx, :] = (
            emb_sifted.A if issparse(emb_sifted) else emb_sifted
        )

        key_added, attr = _save_res(
            adata=self.adata,
            embedding=emb_sifted_full,
            type_emb=type_emb,
            type_res="sift",
            embedding_key=embedding_key,
            key_added=key_added,
        )
        self._stage = Stage.SIFTED

        if type_emb == "X":
            logger.info(
                "The data is `SiFTed`!\n"
                "The filtered embedding is stored in `adata.X`\n"
                "    Finish"
            )
        else:
            logger.info(
                f"The data is `SiFTed`!\n"
                f"The filtered embedding is stored in `adata.{attr}[{key_added!r}]`\n"
                f"    Finish"
            )
        if self._copy:
            return self.adata
        return

    def plot_kernel(
        self, save_path: Optional[os.PathLike] = None, show: bool = True, **kwargs: Any
    ) -> Optional[plt.Axes]:
        """Visualize the cell-cell similarity kernel.

        Parameters
        ----------
        save_path
            Path where to save the figure.
        show
            If `False`, return :class:`matplotlib.axes.Axes`.
        kwargs
            Additional keyword arguments to :func:`sift._utils.plot_kernel()`.

        Returns
        -------
        The axes object, if ``show = False``, otherwise `None`.
        """
        kernel_containers = kwargs.pop("kernel_containers", self._kernel_containers)
        groupby = kwargs.pop("groupby", None)
        if self._kernel_metric == KernelType.MAPPING:
            groupby = self._kernel_key
        return _plot_kernel(
            kernel=self.materialize_kernel(
                kernel_containers=kernel_containers, to_cpu=True
            ),
            adata=self.adata,
            src_key=self._src_key,
            tgt_key=self._tgt_key,
            groupby=groupby,
            save_path=save_path,
            show=show,
            **kwargs,
        )


def sifter(
    adata: AnnData,
    kernel_key: Optional[str] = None,
    metric: st.Kernel_t = KernelType.RBF,
    src_key: Optional[str] = None,
    tgt_key: Optional[str] = None,
    knn_key: Optional[str] = None,
    batch_key: Optional[str] = None,
    embedding_key: Optional[str] = None,
    use_raw: Optional[bool] = False,
    pseudocount: bool = False,
    key_added: Optional[str] = None,
    copy: bool = False,
    **kwargs: Any,
) -> AnnData:
    """Perform filtering, subtract the projection of the kernel from the expression.

    Parameters
    ----------
    adata
        Annotated data object.
    kernel_key
        Key in :class:`anndata.AnnData` where information for distance is stored.
    metric
        Metric to use.
    src_key
        Mask for the source space of the kernel.
    tgt_key
        Mask for the target space of the kernel.
    knn_key
        Key in :attr:`anndata.AnnData.obs` for k-NN masking.
    batch_key
        Key in :attr:`anndata.AnnData.obs` for batch masking.
    embedding_key
        Key in :attr:`anndata.AnnData.obsm` for which to compute the projection on.
        If `None`, use :attr:`anndata.AnnData.X`.
    use_raw
        Whether to use :attr:`anndata.AnnData.raw`, if present.
    pseudocount
        If `True`, add a pseudocount to filtered expression to avoid negative values.
    key_added
        Key used when saving the result. If `None`, use ``'{embedding_key}_sift'``.
        The result is saved to :attr:`anndata.AnnData.X`, :attr:`anndata.AnnData.layers` or
        :attr:`anndata.AnnData.obsm`, depending on the ``embedding_key``.
    copy
        Whether to modify ``adata`` inplace.
    kwargs
        Additional keyword arguments [n_neighbors, precomputed_kernel].

    Returns
    -------
    The filtered data matrix.
    """
    return SiFT(
        adata=adata,
        kernel_key=kernel_key,
        metric=metric,
        knn_key=knn_key,
        batch_key=batch_key,
        src_key=src_key,
        tgt_key=tgt_key,
        copy=copy,
        **kwargs,
    ).filter(
        embedding_key=embedding_key,
        use_raw=use_raw,
        pseudocount=pseudocount,
        key_added=key_added,
    )



import os
import random
import string
from pathlib import Path
from typing import Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from anndata import AnnData
from scipy import ndimage
from scipy.sparse import csr_matrix, spmatrix

from sift._constants import UNS_KEY, KernelType

__all__ = ["plot_kernel"]


def _knn_kernel_base(
    adata: Optional[AnnData] = None,
    kernel_key: Optional[str] = None,
    uns_key: Optional[str] = None,
    tgt_key: Optional[str] = None,
    **kwargs,
) -> str:

    n_neighbors = kwargs.pop("n_neighbors", 15)
    batch_key = kwargs.pop("knn_batch_key", None)
    key_added = kwargs.pop("knn_key_added", "sift")
    use_tgt = kwargs.pop("use_tgt", False)
    use_bbknn = kwargs.pop("use_bbknn", True)
    epsilon = 1e-6

    if (tgt_key is not None) and use_tgt:
        from sklearn.neighbors import NearestNeighbors

        tgt_idx = _get_mask_key(adata=adata, key=tgt_key)
        tgt_idx_idx = np.argwhere(tgt_idx.values).flatten()
        neigh = NearestNeighbors(n_neighbors=n_neighbors)
        container = _get_key(adata, kernel_key)
        neigh.fit(container[tgt_idx, :])
        distances_, neighbors_ = neigh.kneighbors(container)
        rows = (
            np.zeros((adata.n_obs, n_neighbors)) + np.arange(adata.n_obs)[:, np.newaxis]
        ).flatten()
        cols = tgt_idx_idx[neighbors_.flatten()]
        distances_ = (
            1 - (distances_ / distances_.max(1)[:, np.newaxis])
        ).flatten() + epsilon
        adata.obsm[f"{key_added}_connectivities"] = csr_matrix(
            (distances_, (rows, cols)), shape=(adata.n_obs, adata.n_obs)
        )

    else:
        try:
            from bbknn import bbknn

            BBKNN = True
        except ImportError:
            BBKNN = False
        BBKNN = BBKNN and use_bbknn
        if BBKNN:
            if batch_key is None:  # add dummy batch
                batch_key = "".join(random.sample(string.ascii_lowercase, 10))
                adata.obs[batch_key] = 1
                sc.external.pp.bbknn(
                    adata,
                    batch_key=batch_key,
                    use_rep=kernel_key,
                    neighbors_within_batch=n_neighbors,
                )

                del adata.obs[batch_key]
            else:
                neighbors_within_batch = int(
                    n_neighbors / len(adata.obs[batch_key].unique())
                )
                sc.external.pp.bbknn(
                    adata,
                    batch_key=batch_key,
                    use_rep=kernel_key,
                    neighbors_within_batch=neighbors_within_batch,
                )
                adata.uns[uns_key]["batch_key"] = batch_key

            adata.obsp[f"{key_added}_connectivities"] = adata.obsp[
                "connectivities"
            ].copy()

            del adata.obsp["connectivities"]
            del adata.obsp["distances"]
            del adata.uns["neighbors"]
        else:
            sc.pp.neighbors(
                adata, n_neighbors=n_neighbors, use_rep=kernel_key, key_added=key_added
            )
    kernel_key = f"{key_added}_connectivities"
    return kernel_key


def _kernel_base(
    adata: Optional[AnnData] = None,
    kernel_key: Optional[str] = None,
    metric: Optional[str] = KernelType.RBF,
    precomputed_kernel: Optional[np.ndarray] = None,
    uns_key: Optional[str] = None,
    tgt_key: Optional[str] = None,
    **kwargs,
) -> np.ndarray:
    adata.uns[uns_key] = {}
    if precomputed_kernel is not None:
        adata.uns[uns_key]["metric"] = KernelType.PRECOMPUTED
        container = precomputed_kernel

    else:
        if metric == KernelType.KNN:
            kernel_key = _knn_kernel_base(
                adata=adata,
                kernel_key=kernel_key,
                uns_key=uns_key,
                tgt_key=tgt_key,
                **kwargs,
            )

        adata.uns[uns_key]["kernel_key"] = kernel_key
        adata.uns[uns_key]["metric"] = metric.name
        container = _get_key(adata, kernel_key)
    if container.ndim == 1:
        if metric == KernelType.MAPPING:
            if container.dtype.name != "category":
                n_bins = kwargs.pop("n_bins", 15)
                container = pd.cut(container, bins=n_bins)
                adata.uns[uns_key]["n_bins"] = n_bins
                adata.uns[uns_key]["bins"] = container
            container = pd.get_dummies(container).values
        else:
            container = container[:, np.newaxis]
    return container


def _prepare_kernel_base(
    adata: AnnData,
    container: np.ndarray,
    metric: Optional[str] = KernelType.RBF,
    src_key: Optional[str] = None,
    tgt_key: Optional[str] = None,
    uns_key: Optional[str] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    src_idx = np.arange(container.shape[0])
    tgt_idx = np.arange(container.shape[0])
    if src_key is not None:
        src_idx = _get_mask_idx(adata=adata, key=src_key)
        adata.uns[uns_key]["src_key"] = src_key
    if tgt_key is not None:
        tgt_idx = _get_mask_idx(adata=adata, key=tgt_key)
        adata.uns[uns_key]["tgt_key"] = tgt_key

    x = container[src_idx, :]
    y = container[tgt_idx, :]
    if metric == KernelType.KNN or metric == KernelType.PRECOMPUTED:
        x = x[:, tgt_idx]
        y = None

    return x, y


def _prepare_kernel_container(
    adata: Optional[AnnData] = None,
    kernel_key: Optional[str] = None,
    metric: Optional[str] = KernelType.RBF,
    precomputed_kernel: Optional[np.ndarray] = None,
    src_key: Optional[str] = None,
    tgt_key: Optional[str] = None,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    uns_key = kwargs.pop("uns_key", UNS_KEY)
    container = _kernel_base(
        adata=adata,
        kernel_key=kernel_key,
        metric=metric,
        precomputed_kernel=precomputed_kernel,
        uns_key=uns_key,
        tgt_key=tgt_key,
        **kwargs,
    )

    return _prepare_kernel_base(
        adata=adata,
        container=container,
        metric=metric,
        src_key=src_key,
        tgt_key=tgt_key,
        uns_key=uns_key,
    )


def _get_embedding(
    adata: AnnData,
    embedding_key: Optional[str] = None,
) -> Tuple[Union[np.ndarray, spmatrix], str, str]:
    if embedding_key is not None:
        emb, type_emb = _get_basis(adata, embedding_key)
    else:
        embedding_key = "X"
        type_emb = "X"
        emb = adata.X

    return emb, type_emb, embedding_key


def _save_res(
    adata: AnnData,
    embedding: np.ndarray,
    type_emb: str,
    type_res: Optional[str] = "sift",
    embedding_key: Optional[str] = None,
    key_added: Optional[str] = None,
) -> Tuple[str, str]:
    # save projection\filtered
    key_added = key_added if key_added is not None else f"{embedding_key}_{type_res}"

    if type_emb == "X" and type_res == "sift":
        adata.X = embedding.copy()
        attr = "x"
    elif type_emb == "layer" or type_emb == "X":
        adata.layers[key_added] = embedding.copy()
        attr = "layers"
    elif type_emb == "obsm" or type_emb == "X_obsm":
        adata.obsm[key_added] = embedding.copy()
        attr = "obsm"

    return key_added, attr


def _get_basis(adata: AnnData, basis: str) -> Tuple[Union[np.ndarray, spmatrix], str]:
    if basis == "X":
        return adata.X, "X"
    if f"X_{basis}" in adata.obsm:
        return adata.obsm[f"X_{basis}"], "X_obsm"
    elif basis in adata.obsm:
        return adata.obsm[basis], "obsm"
    elif basis in adata.layers:
        return adata.layers[basis], "layer"
    else:
        raise KeyError(
            f"Unable to find a basis in `adata.obsm['X_{basis}']`,"
            f" `adata.obsm[{basis!r}]` or adata.layer[{basis!r}."
        ) from None


def _get_key(adata: AnnData, key: str) -> np.ndarray:
    if key in adata.obsm:
        container = adata.obsm[key]
    elif f"X_{key}" in adata.obsm:
        container = adata.obsm[f"X_{key}"]
    elif key in adata.obs:
        container = adata.obs[key]
    elif key in adata.obsp:
        container = adata.obsp[key]
    elif key in adata.var:  # if key defines a subset of marker genes
        container = adata.X[:, adata.var[key]]
    else:
        raise KeyError(
            f"Unable to find a basis in:"
            f" `adata.obsm['X_{key}']` or `adata.obsm[{key!r}]`"
            f" or `adata.obs[{key!r}]` "
            f"or `adata.obsm['X_{key}']` or `adata.obsp[{key!r}]`"
            f" or `adata.var[{key!r}]`."
        ) from None

    return container


def _get_mask_key(adata: AnnData, key: str) -> np.ndarray:
    if key in adata.obs:
        container = adata.obs[key]
    elif key in adata.obsm:
        container = adata.obsm[key]
    elif key in adata.var:  # if key defines a subset of marker genes
        container = adata.var[key]
    else:
        raise KeyError(
            f"Unable to find a basis in:"
            f" `adata.obsm[{key!r}]` or `adata.obs[{key!r}]` "
            f"or `adata.var[{key!r}]`."
        ) from None
    if container.dtype != "bool":
        raise KeyError(
            f"Masking key must be of 'bool' type," f" got {container.dtype} instead."
        ) from None
    return container


def _get_mask_idx(adata: AnnData, key: str) -> np.ndarray:
    return np.where(_get_mask_key(adata=adata, key=key))[0]


def _maybe_create_dir(dirpath: Union[str, os.PathLike]) -> None:
    """
    Try creating a directory if it does not already exist.
    Parameters
    ----------
    dirpath
        Path of the directory to create.
    Returns
    -------
    None
        Nothing, just creates a directory if it doesn't exist.
    """

    if not os.path.exists(dirpath) or not os.path.isdir(dirpath):
        try:
            os.makedirs(dirpath, exist_ok=True)
        except OSError:
            pass


def plot_kernel(
    kernel: np.ndarray,
    adata: Optional[AnnData] = None,
    src_key: Optional[str] = None,
    tgt_key: Optional[str] = None,
    groupby: Optional[Union[str, Sequence[str]]] = None,
    save_path: Optional[Union[str, Path]] = None,
    show: bool = True,
    **kwargs,
) -> Optional[plt.Axes]:
    """
    Visualize the cell-cell similarity kernel.
    Parameters
    ----------
    kernel
        kernel to plot
    adata
         Optional, reference adata
    src_key
        Optional, mask for source space of kernel
    tgt_key
        Optional, mask for target space of kernel
    groupby
        the key of the observation grouping to consider.
    save_path
        path and fig_name to save the fig
    show
        If `False`, return :class:`matplotlib.pyplot.Axes`.
    **kwargs
        additional plotting arguments
    Returns
    -------
    The axes object, if ``show = False``.
    """
    interpolation = kwargs.pop("interpolation", True)
    if interpolation:
        kernel = ndimage.gaussian_filter(kernel, sigma=3)
    vmax = kwargs.pop("vmax", np.quantile(kernel, 0.9))
    vmax = vmax if vmax > 0 else np.max(kernel)
    color_palette = kwargs.pop("color_palette", "deep")
    cmap = kwargs.pop("cmap", "Blues")
    ncol = kwargs.pop("ncol", 2)
    figsize = kwargs.pop("figsize", None)
    fontsize = kwargs.pop("fontsize", 20)
    if groupby is not None:
        src_idx = (
            _get_mask_idx(adata=adata, key=src_key)
            if src_key is not None
            else np.arange(adata.n_obs)
        )
        tgt_idx = (
            _get_mask_idx(adata=adata, key=tgt_key)
            if tgt_key is not None
            else np.arange(adata.n_obs)
        )
        if adata is None:
            raise ValueError("cannot use groupby without anndata reference.")

        if groupby in adata.obs:
            groups = adata.obs[groupby]
        elif groupby in adata.uns[UNS_KEY]:
            groups = adata.uns[UNS_KEY][groupby]
        else:
            raise ValueError(
                "cannot find groupby in `anndata.obs` " 'or `anndata.uns["kernel"]`.'
            )

        if groups.dtype.name == "category":
            labels = groups.cat.categories.astype(str)
            df = pd.DataFrame(
                groups.astype(str), index=adata.obs_names, columns=[groupby]
            )
        else:
            labels = groups.unique()
            df = pd.DataFrame(
                groups.astype(str), index=adata.obs_names, columns=[groupby]
            )

        label_pal = sns.color_palette(color_palette, labels.size)
        label_lut = dict(zip(map(str, labels), label_pal))
        row_colors = pd.Series(
            df[groupby].iloc[src_idx], index=adata.obs_names[src_idx], name=groupby
        ).map(label_lut)
        col_colors = pd.Series(
            df[groupby].iloc[tgt_idx], index=adata.obs_names[tgt_idx], name=groupby
        ).map(label_lut)

        g = sns.clustermap(
            kernel,
            row_cluster=False,
            col_cluster=False,
            standard_scale=False,
            col_colors=col_colors.to_numpy(),
            row_colors=row_colors.to_numpy(),
            linewidths=0,
            xticklabels=[],
            yticklabels=[],
            cmap=cmap,
            vmin=0,
            vmax=vmax,
            figsize=figsize,
            **kwargs,
        )
        if adata.obs[groupby].dtype.name == "category":
            for label in labels:
                g.ax_col_dendrogram.bar(
                    0, 0, color=label_lut[label], label=label, linewidth=0
                )

            g.ax_col_dendrogram.legend(
                title=groupby,
                loc="center",
                ncol=ncol,
                bbox_to_anchor=(0.47, 0.9),
                bbox_transform=plt.gcf().transFigure,
            )
    else:
        g = sns.clustermap(
            kernel,
            row_cluster=False,
            col_cluster=False,
            linewidths=0,
            xticklabels=[],
            yticklabels=[],
            cmap=cmap,
            vmin=0,
            vmax=vmax,
            figsize=figsize,
        )

    g.cax.set_position([0.05, 0.2, 0.03, 0.45])
    ax = g.ax_heatmap
    ax.set_xlabel("cells", fontsize=fontsize)
    ax.set_ylabel("cells", fontsize=fontsize)

    if save_path is not None:
        g.fig.savefig(save_path, bbox_inches="tight", transparent=True)

    if not show:
        return g.cax



from enum import Enum

__all__ = ["Backend", "KernelType", "Stage", "MAX_SRC_SIZE", "BATCH_SIZE", "UNS_KEY"]


class Backend(str, Enum):
    TORCH = "torch"
    NUMPY = "numpy"


class KernelType(str, Enum):
    PRECOMPUTED = "precomputed"
    KNN = "knn"
    MAPPING = "mapping"
    RBF = "rbf"
    LAPLACIAN = "laplacian"


class Stage(str, Enum):
    INITIALIZED = "initialized"
    SIFTED = "sifted"


MAX_SRC_SIZE = 400_000
BATCH_SIZE = 100_000
UNS_KEY = "kernel"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sift']

package_data = \
{'': ['*']}

install_requires = \
['anndata>=0.8',
 'matplotlib>=3.3',
 'numpy>=1.15',
 'pandas>=1.3',
 'pykeops>=2.0',
 'rich>=10',
 'scanpy>=1.7',
 'scipy>=1.5.0',
 'seaborn>=0.10']

extras_require = \
{'dev': ['pre-commit>=2.10'],
 'docs': ['sphinx>=4',
          'furo>=2022.04.07',
          'nbsphinx>=0.8',
          'sphinx-autodoc-typehints>=1.12'],
 'gpu': ['torch>=1.8.0'],
 'test': ['pytest>=7']}

setup_kwargs = {
    'name': 'sift-sc',
    'version': '0.1.0',
    'description': 'Biological signal filtering in single-cell data.',
    'long_description': 'SiFT - Biological signal filtering in single-cell data\n======================================================\n\n.. image:: https://raw.githubusercontent.com/nitzanlab/sift-sc/main/docs/_static/img/sift_gc.png\n    :width: 200px\n    :align: center\n    :alt: SiFT logo\n\nSignal FilTering is a tool for uncovering hidden biological processes in single-cell data.\nIt can be applied to a wide range of tasks, from the removal of unwanted variation as a pre-processing step,\nthrough revealing hidden biological structure by utilizing prior knowledge with respect to existing signal,\nto uncovering trajectories of interest using reference data to remove unwanted variation.\n\n.. image:: https://raw.githubusercontent.com/nitzanlab/sift-sc/main/docs/_static/img/sift_abs.png\n    :width: 600px\n    :align: center\n    :alt: SiFT pipeline\n\nVisit our `documentation`_ for installation, tutorials, examples and more.\n\nManuscript\n----------\nPlease see our manuscript `Zoe Piran and Mor Nitzan (2022)`_.\n\nInstallation\n------------\nInstall SiFT via PyPI by running::\n\n    pip install sift-sc\n\n.. _documentation: https://sift-sc.readthedocs.io/\n.. _Zoe Piran and Mor Nitzan (2022): https://github.com/nitzanlab/sift-sc\n',
    'author': 'Zoe Piran',
    'author_email': 'zoe.piran@mail.huji.ac.il',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nitzanlab/sift-sc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import importlib
from types import ModuleType
from typing import Optional, Union

import numpy as np
import scipy.sparse as sp

import sift._types as st
from sift._constants import Backend

__all__ = [
    "is_torch_available",
    "is_torch_tensor",
    "is_numpy_tensor",
    "get_device",
    "get_dtype",
    "get_backend",
    "assert_backend_available",
]


def _safe_import(modname: str, *, raise_exc: bool = False) -> Optional[ModuleType]:
    try:
        return importlib.import_module(modname)
    except ImportError:
        if not raise_exc:
            return None
        raise ImportError("TODO") from None


def is_torch_available() -> bool:
    return _safe_import("torch", raise_exc=False) is not None


def is_numpy_tensor(x: st.ArrayLike, allow_sparse: bool = False) -> bool:
    return isinstance(x, np.ndarray) or (allow_sparse and sp.issparse(x))


def is_torch_tensor(x: st.ArrayLike) -> bool:
    torch = _safe_import("torch")
    return torch.is_tensor(x) if torch else False


def get_device(backend: Backend) -> Optional[st.Device_t]:
    if backend == Backend.NUMPY:
        return None
    if backend == Backend.TORCH:
        import torch

        return torch.device("cuda" if torch.cuda.is_available() else "cpu")
    raise NotImplementedError(backend)


def get_dtype(backend: Backend) -> st.DtypeLike:
    if backend == Backend.NUMPY:
        return np.float32
    if backend == Backend.TORCH:
        import torch

        return torch.float32
    raise NotImplementedError(backend)


def get_backend(backend: Optional[Union[str, Backend]]) -> Backend:
    if backend is None:
        return Backend.TORCH if is_torch_available() else Backend.NUMPY
    backend = Backend(backend)
    assert_backend_available(backend)
    return backend


def assert_backend_available(backend: Backend) -> None:
    if backend == Backend.NUMPY:
        return
    if backend == Backend.TORCH:
        assert is_torch_available(), f"Install `{backend}` first."
